//-----------------------------------------------------------------------------
// Name: GamePlay.cpp
//
// Desc: Code to implement the game play.
//
// Copyright (c) 1998-1999 Microsoft Corp.  All rights reserved.
// Ported by Vadim Bodrov, TMT Development Corp.
//-----------------------------------------------------------------------------

unit GamePlay;

interface

uses Windows, Messages, MMSystem, Strings, Math;

{$i resource.inc}

//-----------------------------------------------------------------------------
// Game global variables
//-----------------------------------------------------------------------------
const
    NUM_ROWS      = 8;
    WM_GAMEOVER   = WM_USER+1;
var
    g_hbmPlayerLeft,    g_hbmPlayerRight, g_hbmPlayerExploding: HBitmap;
    g_hbmWeaponL1,      g_hbmWeaponL2,    g_hbmWeaponR1, g_hbmWeaponR2: HBitmap;
    g_hbmBarrierLeft,   g_hbmBarrierRight: HBitmap;
    g_hbmExplosion1,    g_hbmExplosion2: HBitmap;
    g_hbmShieldPowerUp, g_hbmTimeStopPowerUp: HBitmap;
    g_hbmBackBuffer: HBitmap;

type Weapon = record
    hbm:         HBitmap;
    fX:          Double;
    iVelocity:   Longint;
    x, y, w, h:  Longint;
    iExploding:  Longint;
    iPointValue: Longint;
    bVisible:    Boolean;
    iPowerUp:    Longint;
end;

const
    NUM_WEAPONS = 5;

var
    g_Weapons: array[0..NUM_WEAPONS-1] of Weapon;

const
    NUM_STARS   = 300;

var
    g_Stars: array[0..NUM_STARS-1] of TPoint;

    g_fSpaceshipY: Double          := 0.0;
    g_iSpaceshipY: Longint         := 0;
    g_iSpaceshipX: Longint         := 0;
    g_bSpaceshipFire: Boolean      := FALSE;
    g_bSpaceshipFacesLeft: Boolean := TRUE;
    g_fSpeedScale: Double          := 1.0;
    g_fFireRechargeTime: Double    := 0.0;
    g_fShieldTime: Double          := 1.0;
    g_fTimeStopTime: Double        := 1.0;
    g_dwScore: DWORD               := 0;
    g_bGameOver: Boolean           := FALSE;
    g_fLastTime: Double;
    g_fTimeLapse: Double;
    g_wGameWidth: Word;
    g_wGameHeight: Word;
    g_fCurrentTime: Double;         // Game time in seconds

procedure MovePlayer( bLeft, bRight, bUp, bDown, bFire: Boolean );
procedure AnimateScene( Wnd: HWND );
procedure RenderScene( Wnd: HWND );
procedure PauseGame( bPause: Boolean );
procedure StartNewGame;
procedure InitializeGameObjects(hInst: THandle; Wnd: HWND );
procedure DeleteGameObjects;

implementation

//-----------------------------------------------------------------------------
// Inline helper functions
//-----------------------------------------------------------------------------
function LoadBitmapFromResource(hInst: THandle; dwID: DWORD): HBitMap;
begin
    Result := LoadImage(hInst, MAKEINTRESOURCE(dwID),
                               IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );
end;




//-----------------------------------------------------------------------------
// Name: InitializeGameObjects()
// Desc: Creates the bitmaps for the game. Also, initializes the weapons.
//-----------------------------------------------------------------------------
procedure InitializeGameObjects(hInst: THandle; Wnd: HWND );
var
    rc: TRect;
    DC: HDC;
begin
    g_hbmPlayerLeft      := LoadBitmapFromResource( hInst, IDB_PLAYER_LEFT );
    g_hbmPlayerRight     := LoadBitmapFromResource( hInst, IDB_PLAYER_RIGHT );
    g_hbmPlayerExploding := LoadBitmapFromResource( hInst, IDB_PLAYER_EXPLODING );
    g_hbmBarrierLeft     := LoadBitmapFromResource( hInst, IDB_BARRIER_LEFT );
    g_hbmBarrierRight    := LoadBitmapFromResource( hInst, IDB_BARRIER_RIGHT );
    g_hbmWeaponL1        := LoadBitmapFromResource( hInst, IDB_WEAPON_L1 );
    g_hbmWeaponL2        := LoadBitmapFromResource( hInst, IDB_WEAPON_L2 );
    g_hbmWeaponR1        := LoadBitmapFromResource( hInst, IDB_WEAPON_R1 );
    g_hbmWeaponR2        := LoadBitmapFromResource( hInst, IDB_WEAPON_R2 );
    g_hbmExplosion1      := LoadBitmapFromResource( hInst, IDB_EXPLOSION1 );
    g_hbmExplosion2      := LoadBitmapFromResource( hInst, IDB_EXPLOSION2 );
    g_hbmShieldPowerUp   := LoadBitmapFromResource( hInst, IDB_SHIELD_POWERUP );
    g_hbmTimeStopPowerUp := LoadBitmapFromResource( hInst, IDB_TIMESTOP_POWERUP );

    // Get the game window's dimensions
    GetClientRect( Wnd, rc );
    g_wGameWidth  := WORD(rc.right - rc.left);
    g_wGameHeight := WORD(rc.bottom - rc.top);
    g_iSpaceshipX := g_wGameWidth div 2;


    // Create a back buffer for rendering
    DC := GetDC( Wnd );
    g_hbmBackBuffer := CreateCompatibleBitmap( DC, g_wGameWidth, g_wGameHeight );
    ReleaseDC( Wnd, DC );
end;




//-----------------------------------------------------------------------------
// Name: DeleteGameObjects()
// Desc: Delete objects used by the game
//-----------------------------------------------------------------------------
procedure DeleteGameObjects;
begin
    DeleteObject( g_hbmPlayerLeft );
    DeleteObject( g_hbmPlayerRight );
    DeleteObject( g_hbmPlayerExploding );
    DeleteObject( g_hbmBarrierLeft );
    DeleteObject( g_hbmBarrierRight );
    DeleteObject( g_hbmWeaponL1 );
    DeleteObject( g_hbmWeaponL2 );
    DeleteObject( g_hbmWeaponR1 );
    DeleteObject( g_hbmWeaponR2 );
    DeleteObject( g_hbmExplosion1 );
    DeleteObject( g_hbmExplosion2 );
    DeleteObject( g_hbmShieldPowerUp );
    DeleteObject( g_hbmTimeStopPowerUp );
    DeleteObject( g_hbmBackBuffer );
end;




//-----------------------------------------------------------------------------
// Name: StartNewGame()
// Desc: Resets all game variables for a new game.
//-----------------------------------------------------------------------------
procedure StartNewGame;
var
    i: Longint;
begin
    // Initialize game variables
    ZeroMemory( @g_Weapons, NUM_WEAPONS * sizeof(Weapon) );
    g_fSpaceshipY         := 0.0;
    g_iSpaceshipY         := 0;
    g_bSpaceshipFire      := FALSE;
    g_bSpaceshipFacesLeft := TRUE;
    g_fFireRechargeTime   := 0.0;
    g_fSpeedScale         := 1.0;
    g_fShieldTime         := 0.0;
    g_fTimeStopTime       := 0.0;
    g_dwScore             := 0;
    g_bGameOver           := FALSE;
    g_fLastTime           := timeGetTime * 0.001;

    // Generate some stars for the background
    Randomize;
    for i := 0 to NUM_STARS - 1 do
    begin
        g_Stars[i].x := random(300);
        g_Stars[i].y := random(350);
    end;
end;




//-----------------------------------------------------------------------------
// Name: PauseGame()
// Desc: Pauses or unpauses the game
//-----------------------------------------------------------------------------
procedure PauseGame( bPause: Boolean );
begin
    // Sync up to the system time if unpausing the game
    if not bPause then
        g_fLastTime := timeGetTime * 0.001;
end;



//-----------------------------------------------------------------------------
// Name: MovePlayer()
// Desc: Move the players ship
//-----------------------------------------------------------------------------
procedure MovePlayer( bLeft, bRight, bUp, bDown, bFire: Boolean );
begin
    if g_bGameOver then
        exit;

    // Determine the time (in seconds) since the last frame
    g_fTimeLapse := g_fCurrentTime - g_fLastTime;
    g_fLastTime  := g_fCurrentTime;

    // Update the variables for the player's ship
    if bLeft  then  g_bSpaceshipFacesLeft := TRUE;
    if bRight then  g_bSpaceshipFacesLeft := FALSE;
    if bUp    then  g_fSpaceshipY -:= g_fTimeLapse * 100.0;
    if bDown  then  g_fSpaceshipY +:= g_fTimeLapse * 100.0;

    // Bounds checking for the spaceship
    if g_fSpaceshipY < 20.0 then
        g_fSpaceshipY := 20.0;
    if g_fSpaceshipY > (g_wGameHeight - 20.0) then
        g_fSpaceshipY := g_wGameHeight - 20.0;
    g_iSpaceshipY := trunc(g_fSpaceshipY);

    if bFire then
        if (g_fFireRechargeTime - g_fCurrentTime) < 0.0 then
            g_bSpaceshipFire := TRUE;
end;


var
    PowerUpCounter: Longint := 0;

//-----------------------------------------------------------------------------
// Name: AnimateScene()
// Desc: Update the objects in the scene.
//-----------------------------------------------------------------------------
procedure AnimateScene( Wnd: HWND );
var
    i, j, b: Longint;
begin
    // Update the weapons
    for i :=0 to NUM_WEAPONS - 1 do
    begin
        if g_Weapons[i].bVisible then
        begin
            // Update weapons (unless time is stopped, or game is over)
            if not g_bGameOver then
                if (g_fTimeStopTime - g_fCurrentTime) < 0.0 then
                    g_Weapons[i].fX +:= g_fSpeedScale * g_Weapons[i].iVelocity * g_fTimeLapse;

            if g_Weapons[i].iExploding <> 0 then
            begin
                if (g_Weapons[i].iExploding mod 2) <> 0 then
                    g_Weapons[i].hbm := g_hbmExplosion2
                else
                    g_Weapons[i].hbm := g_hbmExplosion1;
                g_Weapons[i].iExploding -:= 1;
                if g_Weapons[i].iExploding = 0 then
                    g_Weapons[i].bVisible := FALSE;
            end;

            // Check if the weapon went off the screen
            if trunc(g_Weapons[i].fX) > g_wGameWidth then
                g_Weapons[i].bVisible := FALSE;
            if trunc(g_Weapons[i].fX + g_Weapons[i].w) < 0 then
                g_Weapons[i].bVisible := FALSE;
        end;
    end;

    // Insert new weapons
    for i := 0 to NUM_WEAPONS - 1 do
    begin
        if not g_Weapons[i].bVisible then
        begin
            // Select a new velocity
            g_Weapons[i].iVelocity := 35*((timeGetTime mod 7) - 3);
            g_Weapons[i].iPointValue := abs((timeGetTime mod 7) - 3);

            if (g_Weapons[i].iVelocity) <> 0 then
            begin
                g_Weapons[i].bVisible := TRUE;

                PowerUpCounter := 0;
                PowerUpCounter +:= 1;

                // Assign a bitmap, which may be a powerup, or a left- or
                // right-facing weapon
                if 10 = PowerUpCounter then
                begin
                    g_Weapons[i].iPowerUp    := 1;
                    g_Weapons[i].hbm         := g_hbmShieldPowerUp;
                    g_Weapons[i].w           := 20;
                    g_Weapons[i].iPointValue := 1000;
                end
                else if 20 = PowerUpCounter then
                begin
                    g_Weapons[i].iPowerUp    := 2;
                    g_Weapons[i].hbm         := g_hbmTimeStopPowerUp;
                    g_Weapons[i].w           := 20;
                    g_Weapons[i].iPointValue := 1000;

                    PowerUpCounter := 0; // Reset the powerup counter
                end
                else
                begin
                    g_Weapons[i].iPowerUp := 0;

                    b := timeGetTime mod 2;

                    if b = 0 then
                    begin
                        if g_Weapons[i].iVelocity < 0 then
                            g_Weapons[i].hbm      := g_hbmWeaponR1
                        else
                            g_Weapons[i].hbm      := g_hbmWeaponL1;

                        g_Weapons[i].w            := 96;
                        g_Weapons[i].iPointValue  := g_Weapons[i].iPointValue * 200;
                    end
                    else
                    begin
                        if g_Weapons[i].iVelocity < 0 then
                            g_Weapons[i].hbm      := g_hbmWeaponR2
                        else
                            g_Weapons[i].hbm      := g_hbmWeaponL2;

                        g_Weapons[i].w            :=  48;
                        g_Weapons[i].iPointValue  := g_Weapons[i].iPointValue * 100;
                    end;
                end;

                if g_Weapons[i].iVelocity < 0 then
                    g_Weapons[i].fX := g_wGameWidth
                else
                    g_Weapons[i].fX := -g_Weapons[i].w;

                g_Weapons[i].y := timeGetTime mod NUM_ROWS;
                g_Weapons[i].h := 20;

                for j := 0 to NUM_WEAPONS - 1 do
                begin
                    if i = j then
                        continue;
                    if g_Weapons[j].bVisible = FALSE then
                        continue;
                    if g_Weapons[i].y = g_Weapons[j].y then
                    begin
                        g_Weapons[i].y +:= 3;
                        if g_Weapons[i].y >= NUM_ROWS then
                            g_Weapons[i].y -:= NUM_ROWS;
                        j := -1;
                    end;
                end;
            end;

            // Only insert max of 1 new weapon per frame
            break;
        end;
    end;

    // Check for collision
    for i := 0 to NUM_WEAPONS - 1 do
    begin
        if g_Weapons[i].bVisible then
        declare
        var
            x: Longint := trunc(g_Weapons[i].fX);
            y: Longint := g_Weapons[i].y*40 + 20;
            w: Longint := g_Weapons[i].w;
            h: Longint := g_Weapons[i].h;
        begin
            if ( x > (g_iSpaceshipX + 11) ) or ( (x + w) < (g_iSpaceshipX - 11) ) then
                continue;
            if ( y > (g_iSpaceshipY + 11) ) or ( (y + h) < (g_iSpaceshipY - 11) ) then
                continue;
            if g_Weapons[i].iExploding <> 0 then
                continue;

            // Else, the player hit object
            if g_Weapons[i].iPowerUp = 1 then
            begin
                g_fShieldTime := g_fCurrentTime + 5.0; // Shields last for 5 seconds
                g_dwScore +:= g_Weapons[i].iPointValue;
                g_Weapons[i].bVisible := FALSE;
            end;
            if g_Weapons[i].iPowerUp = 2 then
            begin
                g_fTimeStopTime := g_fCurrentTime + 3.0; // Stop time for 3 seconds
                g_dwScore +:= g_Weapons[i].iPointValue;
                g_Weapons[i].bVisible := FALSE;
            end
            else // Player hit weapon
            begin
                if (g_fShieldTime - g_fCurrentTime) < 0.0 then
                begin
                    g_bGameOver := TRUE;
                    SendMessage( Wnd, WM_GAMEOVER, 0, 0 );
                end
                else
                begin
                    g_Weapons[i].hbm        := g_hbmExplosion1;
                    g_Weapons[i].fX        +:= (w-48)/2;
                    g_Weapons[i].w          := 48;
                    g_Weapons[i].iExploding := 15;
                    g_dwScore              +:= g_Weapons[i].iPointValue;
                end;
            end;
        end;
    end;

    // Handle the laser beam
    if g_bSpaceshipFire then
    declare
    var
        beamX: Longint := g_iSpaceshipX;
        beamY: Longint := g_iSpaceshipY;
    begin
        g_bSpaceshipFire := FALSE;

        if (beamY mod 40) > 20 then // Don't fire at barriers
        begin
            g_fFireRechargeTime := g_fCurrentTime + 0.5;

            // Blow up weapons
            for i := 0 to NUM_WEAPONS - 1 do
            declare
            var
                x: Longint := trunc(g_Weapons[i].fX);
                y: Longint := g_Weapons[i].y*40 + 20;
                w: Longint := g_Weapons[i].w;
                h: Longint := g_Weapons[i].h;
            begin
                if not g_Weapons[i].bVisible then
                    continue;
                if (beamY < y ) or ( beamY >= (y+h) ) then
                    continue;

                if  ( g_bSpaceshipFacesLeft and (x < g_iSpaceshipX )) or
                    ( not g_bSpaceshipFacesLeft and (x > g_iSpaceshipX )) then
                begin
                    g_Weapons[i].hbm := g_hbmExplosion1;
                    g_Weapons[i].fX +:= (w-48)/2;
                    g_Weapons[i].w   := 48;
                    g_Weapons[i].iExploding := 15;
                    g_dwScore +:= g_Weapons[i].iPointValue;
                end;
            end;
        end;
    end;

    // Update the speed factor
    g_fSpeedScale := 1.0 + (g_dwScore/10000.0);
end;




//-----------------------------------------------------------------------------
// Name: RenderScene()
// Desc: Render the scene to a backbuffer, and blt to the HWND.
//-----------------------------------------------------------------------------
procedure RenderScene( Wnd: HWND );
var
    // Create the DC's for rendering the frame into a backbuffer
    dcBackBuffer: HDC  := CreateCompatibleDC( 0 );
    dcBitmap: HDC      := CreateCompatibleDC( 0 );
    // Create a brush and pen
    Pen: HPen          := CreatePen( PS_SOLID, 1, $00ffffff );
    OldPen: HPen       := SelectObject( dcBackBuffer, Pen );
    Brush: HBrush      := CreateSolidBrush( $00000000 );
    OldBrush: HBrush   := SelectObject( dcBackBuffer, Brush );
    // Fill the background
    rc: TRect;
    i: Longint;
    buffer: array[0..79] of Char;
    dcWindow: HDC;
begin
    GetClientRect( Wnd, rc );
    SelectObject( dcBackBuffer, g_hbmBackBuffer );
    FillRect( dcBackBuffer, rc, Brush );

    // Draw stars for the background
    for i := 0 to NUM_STARS - 1 do
    begin
        MoveToEx( dcBackBuffer, g_Stars[i].x,   g_Stars[i].y, nil );
        LineTo(   dcBackBuffer, g_Stars[i].x+1, g_Stars[i].y+1 );
    end;

    // Draw the barriers
    for i := 0 to NUM_ROWS - 1 do
    begin
        SelectObject( dcBitmap, g_hbmBarrierLeft );
        BitBlt( dcBackBuffer, 0, i*40, 140, 20, dcBitmap, 0, 0, SRCCOPY );
        SelectObject( dcBitmap, g_hbmBarrierRight );
        BitBlt( dcBackBuffer, g_wGameWidth-141, i*40, 140, 20, dcBitmap, 0, 0, SRCCOPY );
    end;

    // Draw the weapons
    for i := 0 to NUM_WEAPONS - 1 do
    begin
        if g_Weapons[i].bVisible then
        begin
            if g_Weapons[i].iExploding <> 0 then
            begin
                if (g_Weapons[i].iExploding mod 2) <> 0 then
                    g_Weapons[i].hbm := g_hbmExplosion2
                else
                    g_Weapons[i].hbm := g_hbmExplosion1;
            end;

            declare
            var
                x: Longint := trunc(g_Weapons[i].fX);
                y: Longint := g_Weapons[i].y*40 + 20;
                w: Longint := g_Weapons[i].w;
                h: Longint := g_Weapons[i].h;
            begin
                SelectObject( dcBitmap, g_Weapons[i].hbm );
                BitBlt( dcBackBuffer, x, y, w, h, dcBitmap, 0, 0, SRCCOPY );
            end;
        end;
    end;

    // Draw the laser beam
    if (g_fFireRechargeTime - g_fCurrentTime) > 0.45 then
    begin
        MoveToEx( dcBackBuffer, g_iSpaceshipX, g_iSpaceshipY, nil );

        if g_bSpaceshipFacesLeft then
            LineTo( dcBackBuffer, 0, g_iSpaceshipY )
        else
            LineTo( dcBackBuffer, g_wGameWidth, g_iSpaceshipY );
    end;

    // Draw the spaceship
    if g_bGameOver then
        SelectObject( dcBitmap, g_hbmPlayerExploding )
    else if g_bSpaceshipFacesLeft then
        SelectObject( dcBitmap, g_hbmPlayerLeft )
    else
        SelectObject( dcBitmap, g_hbmPlayerRight );

    BitBlt( dcBackBuffer, g_iSpaceshipX-16, g_iSpaceshipY-16, 32, 32,
            dcBitmap, 0, 0, SRCCOPY );

    // Draw the shield
    if (g_fShieldTime - g_fCurrentTime) > 0.0 then
        Arc( dcBackBuffer, g_iSpaceshipX-16, g_iSpaceshipY-16,
             g_iSpaceshipX+16, g_iSpaceshipY+16, g_iSpaceshipX, g_iSpaceshipY-16,
             g_iSpaceshipX, g_iSpaceshipY-16 );

    // Output the score
    StrPCopy(buffer, ' Score: ' + IntToStr(g_dwScore));
    SetTextColor( dcBackBuffer, RGB(255,255,0) );
    SetBkColor( dcBackBuffer, RGB(0,0,255) );
    ExtTextOut( dcBackBuffer, 0, 2, ETO_OPAQUE, nil, buffer, strlen(buffer), nil );

    // If game is over, output "Game Over" string
    if g_bGameOver then
    declare
    var
        Font: HFont := CreateFont( 48, 0, 0, 0, FW_NORMAL, 0, 0, 0,
                                  ANSI_CHARSET, OUT_DEFAULT_PRECIS,
                                  CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
                                  VARIABLE_PITCH, 'Arial' );
        OldFont: HFont := SelectObject( dcBackBuffer, Font );
    begin
        SetTextColor( dcBackBuffer, RGB(255,255,0) );
        SetBkMode(    dcBackBuffer, TRANSPARENT );
        SetTextAlign( dcBackBuffer, TA_CENTER );

        ExtTextOut(   dcBackBuffer, g_wGameWidth div 2, NUM_ROWS*20-25, 0,
                      nil, 'Game Over', 9, nil );

        SelectObject( dcBackBuffer, OldFont );
        DeleteObject( Font );
    end;

    // Blt the backbuffer to the window
    dcWindow := GetDC( Wnd );
    BitBlt( dcWindow, 0, 0, g_wGameWidth, g_wGameHeight, dcBackBuffer, 0, 0, SRCCOPY );
    ReleaseDC( Wnd, dcWindow );

    // Cleanup used objects
    SelectObject( dcBackBuffer, OldBrush );
    DeleteObject( Brush );
    SelectObject( dcBackBuffer, OldPen );
    DeleteObject( Pen );
    DeleteDC( dcBackBuffer );
    DeleteDC( dcBitmap );
end;


end.