(*==========================================================================
 *
 *  Copyright (C) 1995-1999 Microsoft Corporation. All Rights Reserved.
 *  Ported by Vadim Bodrov, TMT Development Corp.
 *
 *  File:       input.pas
 *  Content:    Input routines for Space Donuts game
 *
 *
 ***************************************************************************)

unit Input;

interface

uses Windows, DXLite;

{$i input.inc}

const
    DONUTS_ICON    = 101;

    DONUTS_MENU    = 1;

    DONUTS_ACCEL   = 1;

    IDC_STARTGAME  = 100;
    IDC_QUIT       = 101;
    IDC_TRAILS     = 102;
    IDC_AUDIO      = 103;
    IDC_FRAMERATE  = 104;

    IDC_DEVICES    = 200;     // 200 through 299

var
    hWndMain: HWND;

function  ReacquireInput: LongBool;
function  PickInputDevice(n: Longint): Boolean;
procedure CleanupInput;
function  InitInput(hInst: THandle; Wnd: HWND): Boolean;

var
    // allocate external variables
    ReadGameInput: function: DWORD;

implementation

(*
 *  We'll use up to the first ten input devices.
 *
 *  c_cpdiFound is the number of found devices.
 *  g_rgpdiFound[0] is the array of found devices.
 *  g_pdevCurrent is the device that we are using for input.
 *)
const
    MAX_DINPUT_DEVICES = 10;

var
    g_cpdevFound: Longint;
    g_rgpdevFound: array[0..MAX_DINPUT_DEVICES - 1] of PPDirectInputDevice2;
    g_pdevCurrent: PPDirectInputDevice2;


(*--------------------------------------------------------------------------
| ReacquireInput
|
| Reacquires the current input device.  If Acquire() returns S_FALSE,
| that means
| that we are already acquired and that DirectInput did nothing.
*-------------------------------------------------------------------------*)
function ReacquireInput: LongBool;
var
    hr: HResult;
begin
    // if we have a current device
    if g_pdevCurrent <> nil then
    begin
       // acquire the device
       hr := g_pdevCurrent^^.Acquire^(g_pdevCurrent);
       if SUCCEEDED(hr) then
       begin
          // acquisition successful
          Result := LONGTRUE;
          exit;
       end else
       begin
          // acquisition failed
          Result := LONGFALSE;
          exit;
       end;
    end else
    begin
       // we don't have a current device
       Result := LONGFALSE;
    end;
end;


var
    dwKeyState: DWORD := 0;

(*--------------------------------------------------------------------------
| ReadKeyboardInput
|
| Requests keyboard data and performs any needed processing.
*-------------------------------------------------------------------------*)
function ReadKeyboardInput: DWORD;
var
   rgKeyData: array[0..KEYBUFSIZE-1] of TDIDeviceObjectData;
   dwEvents:  DWORD;
   dw:        DWORD;
   hRes:      HResult;
begin
   // get data from the keyboard
   dwEvents := KEYBUFSIZE;
   hRes := g_pdevCurrent^^.GetDeviceData^(g_pdevCurrent,
                                          sizeof(TDIDeviceObjectData),
                                          rgKeyData[0], dwEvents, 0);
   if hRes <> DI_OK then
   begin
      // did the read fail because we lost input for some reason?
      // if so, then attempt to reacquire.  If the second acquire
      // fails, then the error from GetDeviceData will be
      // DIERR_NOTACQUIRED, so we won't get stuck an infinite loop.
      if hRes = DIERR_INPUTLOST then ReacquireInput;

      // return the fact that we did not read any data
      Result := 0;
      exit;
   end;

   // process the data
   dw := 0;
   while (dw < dwEvents) do
   begin

      case (rgKeyData[dw].dwOfs) of
      // fire
      DIK_SPACE:
        if (rgKeyData[dw].dwData and $80) <> 0 then
           dwKeyState := dwKeyState or KEY_FIRE
        else
           dwKeyState := dwKeyState and (not KEY_FIRE);

      // stop
      DIK_NUMPAD5:
        if (rgKeyData[dw].dwData and $80) <> 0 then
           dwKeyState := dwKeyState or KEY_STOP
        else
           dwKeyState := dwKeyState and (not KEY_STOP);

      // shield
      DIK_NUMPAD7:
        if (rgKeyData[dw].dwData and $80) <> 0 then
           dwKeyState := dwKeyState or KEY_SHIELD
        else
           dwKeyState := dwKeyState and (not KEY_SHIELD);

      // thrust
      DIK_UP, DIK_NUMPAD8:
        if (rgKeyData[dw].dwData and $80) <> 0 then
           dwKeyState := dwKeyState or KEY_UP
        else
           dwKeyState := dwKeyState and (not KEY_UP);

      // reverse thrust
      DIK_DOWN, DIK_NUMPAD2:
        if (rgKeyData[dw].dwData and $80) <> 0 then
           dwKeyState := dwKeyState or KEY_DOWN
        else
           dwKeyState := dwKeyState and DWORD(not KEY_DOWN);

      // rotate left
      DIK_LEFT, DIK_NUMPAD4:
        if (rgKeyData[dw].dwData and $80) <> 0 then
           dwKeyState := dwKeyState or KEY_LEFT
        else
           dwKeyState := dwKeyState and (not KEY_LEFT);

      // rotate right
      DIK_RIGHT, DIK_NUMPAD6:
        if (rgKeyData[dw].dwData and $80) <> 0 then
           dwKeyState := dwKeyState or KEY_RIGHT
        else
           dwKeyState := dwKeyState and (not KEY_RIGHT);
      end;

      dw +:= 1;
   end;

   Result := dwKeyState;
end;


(*--------------------------------------------------------------------------
| AddInputDevice
|
| Records an input device in the array of found devices.
|
| In addition to stashing it in the array, we also add it to the device
| menu so the user can pick it.
|
*-------------------------------------------------------------------------*)

procedure AddInputDevice(pdev: PPDirectInputDevice; pdi: PDIDeviceInstance);
var
    hRes:  HResult;
    hmenu: HMenu;
begin

    if (g_cpdevFound < MAX_DINPUT_DEVICES) then
    begin

        (*
         *  Convert it to a Device2 so we can Poll() it.
         *)

        hRes := pdev^^.QueryInterface^(
                    pdev, IID_IDirectInputDevice2,
                    g_rgpdevFound[g_cpdevFound]);

        if (SUCCEEDED(hRes)) then
        begin

            (*
             *  Add its description to the menu.
             *)
            hmenu := GetSubMenu(GetMenu(hWndMain), 0);

            InsertMenu(hmenu, g_cpdevFound, MF_BYPOSITION or MF_STRING,
                       IDC_DEVICES + g_cpdevFound,
                       pdi^.tszInstanceName);

            g_cpdevFound +:= 1;
        end;
    end;
end;

(*--------------------------------------------------------------------------
|
| SetDIDwordProperty
|
| Set a DWORD property on a DirectInputDevice.
|
*-------------------------------------------------------------------------*)
function SetDIDwordProperty(pdev: PPDirectInputDevice; guidProperty: PGUID;
    dwObject, dwHow, dwValue: DWORD): HResult;
var
   dipdw: TDIPropDWORD;
begin
   dipdw.diph.dwSize       := sizeof(dipdw);
   dipdw.diph.dwHeaderSize := sizeof(dipdw.diph);
   dipdw.diph.dwObj        := dwObject;
   dipdw.diph.dwHow        := dwHow;
   dipdw.dwData            := dwValue;

   Result := pdev^^.SetProperty^(pdev, guidProperty, dipdw.diph);
end;

(*--------------------------------------------------------------------------
| InitKeyboardInput
|
| Initializes DirectInput for the keyboard.  Creates a keyboard device,
| sets the data format to our custom format, sets the cooperative level and
| adds it to the menu.
|
*-------------------------------------------------------------------------*)
function InitKeyboardInput(pdi: PPDirectInput): Boolean;
var
   pdev: PPDirectInputDevice;
   di:   TDiDeviceInstance;
begin
   // create the DirectInput keyboard device
   if (pdi^^.CreateDevice^(pdi, GUID_SysKeyboard, pdev, nil) <> DI_OK) then
   begin
      OutputDebugString('IDirectInput.CreateDevice FAILED' + #13);
      Result := FALSE;
      exit;
   end;

   // set keyboard data format
   if (pdev^^.SetDataFormat^(pdev, c_dfDIKeyboard) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.SetDataFormat FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := FALSE;
   end;

   // set the cooperative level
   if (pdev^^.SetCooperativeLevel^(pdev, hWndMain,
      DISCL_NONEXCLUSIVE or DISCL_FOREGROUND) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.SetCooperativeLevel FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := FALSE;
   end;

   // set buffer size
   if (SetDIDwordProperty(pdev, DIPROP_BUFFERSIZE, 0, DIPH_DEVICE, KEYBUFSIZE) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.SetProperty(DIPH_DEVICE) FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := FALSE;
   end;

   //
   // Get the name of the primary keyboard so we can add it to the menu.
   //
   di.dwSize := sizeof(di);
   if (pdev^^.GetDeviceInfo^(pdev, di) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.GetDeviceInfo FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := FALSE;
   end;

   //
   // Add it to our list of devices.  If AddInputDevice succeeds,
   // he will do an AddRef.
   //
   AddInputDevice(pdev, @di);
   pdev^^.Release^(pdev);


   Result := TRUE;
end;

(*--------------------------------------------------------------------------
| InitJoystickInput
|
| Initializes DirectInput for an enumerated joystick device.
| Creates the device, device, sets the standard joystick data format,
| sets the cooperative level and adds it to the menu.
|
| If any step fails, just skip the device and go on to the next one.
|
*-------------------------------------------------------------------------*)

function InitJoystickInput conv arg_stdcall (var dinst: TDIDeviceInstanceA; pvRef: Pointer): HResult;
var
   pdi:   PPDirectInput := pvRef;
   pdev:  PPDirectInputDevice;
   diprg: TDIPropRange;
begin
   // create the DirectInput joystick device
   if (pdi^^.CreateDevice^(pdi, dinst.guidInstance, pdev, nil) <> DI_OK) then
   begin
      OutputDebugString('IDirectInput::CreateDevice FAILED' + #13);
      Result := DIENUM_CONTINUE;
      exit;
   end;

   // set joystick data format
   if (pdev^^.SetDataFormat^(pdev, c_dfDIJoystick) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.SetDataFormat FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := DIENUM_CONTINUE;
      exit;
   end;

   // set the cooperative level
   if (pdev^^.SetCooperativeLevel^(pdev, hWndMain,
      DISCL_NONEXCLUSIVE or DISCL_FOREGROUND) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.SetCooperativeLevel FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := DIENUM_CONTINUE;
      exit;
   end;

   // set X-axis range to (-1000 ... +1000)
   // This lets us test against 0 to see which way the stick is pointed.

   diprg.diph.dwSize       := sizeof(diprg);
   diprg.diph.dwHeaderSize := sizeof(diprg.diph);
   diprg.diph.dwObj        := DIJOFS_X;
   diprg.diph.dwHow        := DIPH_BYOFFSET;
   diprg.lMin              := -1000;
   diprg.lMax              := +1000;

   if (pdev^^.SetProperty^(pdev, DIPROP_RANGE, diprg.diph) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.SetProperty(DIPH_RANGE) FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := LONGFALSE;
      exit;
   end;

   //
   // And again for Y-axis range
   //
   diprg.diph.dwObj        := DIJOFS_Y;

   if (pdev^^.SetProperty^(pdev, DIPROP_RANGE, diprg.diph) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.SetProperty(DIPH_RANGE) FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := LONGFALSE;
      exit;
   end;

   // set X axis dead zone to 50% (to avoid accidental turning)
   // Units are ten thousandths, so 50% = 5000/10000.
   if (SetDIDwordProperty(pdev, DIPROP_DEADZONE, DIJOFS_X, DIPH_BYOFFSET, 5000) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice::SetProperty(DIPH_DEADZONE) FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := LONGFALSE;
      exit;
   end;


   // set Y axis dead zone to 50% (to avoid accidental thrust)
   // Units are ten thousandths, so 50% = 5000/10000.
   if (SetDIDwordProperty(pdev, DIPROP_DEADZONE, DIJOFS_Y, DIPH_BYOFFSET, 5000) <> DI_OK) then
   begin
      OutputDebugString('IDirectInputDevice.SetProperty(DIPH_DEADZONE) FAILED' + #13);
      pdev^^.Release^(pdev);
      Result := LONGFALSE;
      exit;
   end;


   //
   // Add it to our list of devices.  If AddInputDevice succeeds,
   // he will do an AddRef.
   //
   AddInputDevice(pdev, @dinst);
   pdev^^.Release^(pdev);

   Result := DIENUM_CONTINUE;
end;


(*--------------------------------------------------------------------------
| ReadJoystickInput
|
| Requests joystick data and performs any needed processing.
|
*-------------------------------------------------------------------------*)
function ReadJoystickInput: DWORD;
var
    dwKeyState: DWORD;
    hRes: HResult;
    js: TDIJoyState;
begin
   // poll the joystick to read the current state
   hRes := g_pdevCurrent^^.Poll^(g_pdevCurrent);

   // get data from the joystick
   hRes := g_pdevCurrent^^.GetDeviceState^(g_pdevCurrent, sizeof(TDIJoyState), js);

   if (hRes <> DI_OK) then
   begin
      // did the read fail because we lost input for some reason?
      // if so, then attempt to reacquire.  If the second acquire
      // fails, then the error from GetDeviceData will be
      // DIERR_NOTACQUIRED, so we won't get stuck an infinite loop.
      if (hRes = DIERR_INPUTLOST) then
         ReacquireInput;

      // return the fact that we did not read any data
      Result := 0;
      exit;
   end;

   //
   // Now study the position of the stick and the buttons.
   //

   dwKeyState := 0;

   if (js.lX < 0) then
      dwKeyState := dwKeyState or KEY_LEFT
   else
   if (js.lX > 0) then
      dwKeyState := dwKeyState or KEY_RIGHT;

   if (js.lY < 0) then
      dwKeyState := dwKeyState or KEY_UP
   else
   if (js.lY > 0) then
      dwKeyState := dwKeyState or KEY_DOWN;

   if (js.rgbButtons[0] and $80) <> 0 then
      dwKeyState := dwKeyState or KEY_FIRE;

   if (js.rgbButtons[1] and $80) <> 0 then
      dwKeyState := dwKeyState or KEY_SHIELD;

   if (js.rgbButtons[2] and $80) <> 0 then
      dwKeyState := dwKeyState or KEY_STOP;

   Result := dwKeyState;
end;


(*--------------------------------------------------------------------------
| PickInputDevice
|
| Make the n'th input device the one that we will use for game play.
|
*-------------------------------------------------------------------------*)
function PickInputDevice(n: Longint): Boolean;
begin
    if (n < g_cpdevFound) then
    begin

        (*
         *  Unacquire the old device.
         *)
        if (g_pdevCurrent <> nil) then
            g_pdevCurrent^^.Unacquire^(g_pdevCurrent);

        (*
         *  Move to the new device.
         *)
        g_pdevCurrent := g_rgpdevFound[n];

        (*
         *  Set ReadGameInput to the appropriate handler.
         *)
        if (n = 0) then
            ReadGameInput := ReadKeyboardInput
        else
            ReadGameInput := ReadJoystickInput;

        CheckMenuRadioItem(GetSubMenu(GetMenu(hWndMain), 0),
                           IDC_DEVICES, IDC_DEVICES + g_cpdevFound - 1,
                           IDC_DEVICES + n, MF_BYCOMMAND);

        ReacquireInput;

        Result := TRUE;
        exit;
    end else
        Result := FALSE;
end;


/*--------------------------------------------------------------------------
| InitInput
|
| Initializes DirectInput for the keyboard and all joysticks.
|
| For each input device, add it to the menu.
|
*-------------------------------------------------------------------------*/

function InitInput(hInst: THandle; Wnd: HWND): Boolean;
var
   pdi: PPDirectInput;
   fRc: Boolean;
begin
   // Note: Joystick support is a DirectX 5.0 feature.
   // Since we also want to run on DirectX 3.0, we will start out
   // with DirectX 3.0 to make sure that at least we get the keyboard.

   // create the DirectInput interface object
   if (DirectInputCreate(hInst, $0300, pdi, nil) <> DI_OK) then
   begin
      OutputDebugString('DirectInputCreate 3.0 FAILED' + #13);
      Result := FALSE;
      exit;
   end;

   fRc := InitKeyboardInput(pdi);
   pdi^^.Release^(pdi);       // Finished with DX 3.0

   if (not fRc) then
   begin
      Result := FALSE;
      exit;
   end;

   // create the DirectInput 5.0 interface object
   if (DirectInputCreate(hInst, DIRECTINPUT_VERSION, pdi, nil) = DI_OK) then
   begin

      //
      // Enumerate the joystick devices.  If it doesn't work, oh well,
      // at least we got the keyboard.
      //

      pdi^^.EnumDevices^(pdi, DIDEVTYPE_JOYSTICK,
          @InitJoystickInput, pdi, DIEDFL_ATTACHEDONLY);

      pdi^^.Release^(pdi);    // Finished with DX 5.0.

   end else
   begin
      OutputDebugString('DirectInputCreate 5.0 FAILED - no joystick support' + #13);
   end;

   // Default device is the keyboard
   PickInputDevice(0);
   // if we get here, we were successful
   Result := TRUE;
end;

(*--------------------------------------------------------------------------
| CleanupInput
|
| Cleans up all DirectInput objects.
*-------------------------------------------------------------------------*)
procedure CleanupInput;
var
    idev: Longint;
begin
   // make sure the device is unacquired
   // it doesn't harm to unacquire a device that isn't acquired

   if (g_pdevCurrent <> nil) then
   {
      IDirectInputDevice_Unacquire(g_pdevCurrent);
   }

   // release all the devices we created
   for idev := 0 to g_cpdevFound - 1 do
   begin
      if (g_rgpdevFound[idev] <> nil) then
      begin
         g_rgpdevFound[idev]^^.Release^(g_rgpdevFound[idev]);
         g_rgpdevFound[idev] := nil;
      end;
   end;

end;


begin
  ReadGameInput := ReadKeyboardInput;
end.